import codecs
import os
from setuptools import setup, find_packages

with codecs.open("README.md", encoding="utf-8") as f:
    readme = f.read()


def get_version_file_path():
    github_actions_path = "/home/runner/work/gitlabform/gitlabform"
    if os.path.isfile(github_actions_path + "/version"):
        return github_actions_path + "/version"
    else:
        return "version"


setup(
    name="gitlabform",
    version=open(get_version_file_path()).read(),
    description='🏗 Specialized "configuration as a code" tool for GitLab projects, groups and more'
    " using hierarchical configuration written in YAML",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://gitlabform.github.io/gitlabform",
    author="Greg Dubicki and Contributors",
    keywords=["cli", "yaml", "gitlab", "configuration-as-code"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Software Development :: Version Control :: Git",
    ],
    packages=find_packages(),
    package_data={"": ["LICENSE", "version", "*.md", "config.yml"]},
    include_package_data=True,
    python_requires=">=3.7.0",
    install_requires=[
        "certifi",  # we want the latest root certs for security
        "requests==2.28.1",
        "types-requests==2.28.11.7",
        "Jinja2==3.1.2",
        "MarkupSafe==2.1.1",
        "ruamel.yaml==0.17.21",
        "luddite==1.0.2",
        "cli-ui==0.17.2",
        "packaging==22.0",
        "mergedeep==1.3.4",
        "yamlpath==3.6.9",
        "ez-yaml==1.2.0",
    ],
    extras_require={
        "test": [
            "pytest==7.2.0",
            "xkcdpass==1.19.3",
            "pre-commit==2.21.0",  # not really for tests, but for development
            "coverage==7.0.1",
            "pytest-cov==4.0.0",
            "deepdiff==6.2.2",
            "pytest-rerunfailures==10.3",
            "cryptography==38.0.4",
            "mypy==0.991",
            "mypy-extensions==0.4.3",
        ],
        "docs": [
            "mkdocs",
            "mkdocs-material",
        ],
    },
    entry_points={
        "console_scripts": [
            "gitlabform=gitlabform.run:run",
        ],
    },
)



import os
from typing import Callable, Optional, Generator, List

import pytest
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend

from gitlabform.gitlab import AccessLevel, GitLab
from tests.acceptance import (
    get_gitlab,
    create_group,
    create_groups,
    delete_groups,
    create_project,
    get_random_name,
    create_users,
    delete_users,
    get_random_password,
)


@pytest.fixture(scope="session")
def gitlab() -> Generator[GitLab, None, None]:
    gl = get_gitlab()
    yield gl  # provide fixture value


@pytest.fixture(scope="class")
def group_and_project(group, project) -> str:
    return f"{group}/{project}"


@pytest.fixture(scope="function")
def group_and_project_for_function(group, project_for_function) -> str:
    return f"{group}/{project_for_function}"


@pytest.fixture(scope="class")
def group() -> Generator[str, None, None]:
    group_name = get_random_name("group")
    create_group(group_name)

    yield group_name

    gl = get_gitlab()
    gl.delete_group(group_name)


@pytest.fixture(scope="function")
def group_for_function():
    group_name = get_random_name("group")
    create_group(group_name)

    yield group_name

    gl = get_gitlab()
    gl.delete_group(group_name)


@pytest.fixture(scope="class")
def other_group():
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    group_name = get_random_name("group")
    create_group(group_name)

    yield group_name

    gl = get_gitlab()
    gl.delete_group(group_name)


@pytest.fixture(scope="class")
def third_group():
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    group_name = get_random_name("group")
    create_group(group_name)

    yield group_name

    gl = get_gitlab()
    gl.delete_group(group_name)


@pytest.fixture(scope="class")
def sub_group(group):
    gl = get_gitlab()
    parent_id = gl.get_group_id_case_insensitive(group)
    group_name = get_random_name("subgroup")
    create_group(group_name, parent_id)

    yield group + "/" + group_name

    gl = get_gitlab()
    gl.delete_group(group + "/" + group_name)


@pytest.fixture(scope="class")
def project(group):
    project_name = get_random_name("project")
    create_project(group, project_name)

    yield project_name

    gl = get_gitlab()
    gl.delete_project(f"{group}/{project_name}")


@pytest.fixture(scope="function")
def project_for_function(group):
    project_name = get_random_name("project")
    create_project(group, project_name)

    yield project_name

    gl = get_gitlab()
    gl.delete_project(f"{group}/{project_name}")


@pytest.fixture(scope="class")
def other_project(group):
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    project_name = get_random_name("project")
    create_project(group, project_name)

    yield project_name

    gl = get_gitlab()
    gl.delete_project(f"{group}/{project_name}")


@pytest.fixture(scope="class")
def groups(users):
    no_of_groups = 4

    group_name_base = get_random_name("group")
    groups = create_groups(group_name_base, no_of_groups)

    yield groups

    delete_groups(group_name_base, no_of_groups)


@pytest.fixture(scope="class")
def users(group):
    no_of_users = 4

    username_base = get_random_name("user")
    users = create_users(username_base, no_of_users)

    yield users

    delete_users(username_base, no_of_users)


@pytest.fixture(scope="class")
def other_users():
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    no_of_users = 4

    username_base = get_random_name("user")
    users = create_users(username_base, no_of_users)

    yield users

    delete_users(username_base, no_of_users)


@pytest.fixture(scope="class")
def branch(gitlab, group_and_project):
    name = get_random_name("branch")
    gitlab.create_branch(group_and_project, name, "main")

    yield name

    gitlab.delete_branch(group_and_project, name)


@pytest.fixture(scope="class")
def other_branch(gitlab, group_and_project):
    # TODO: deduplicate
    name = get_random_name("other_branch")
    gitlab.create_branch(group_and_project, name, "main")

    yield name

    gitlab.delete_branch(group_and_project, name)


class User:
    def __init__(self, name, id):
        self.name = name
        self.id = id


@pytest.fixture(scope="class")
def make_user(
    gitlab, group_and_project
) -> Generator[Callable[[AccessLevel, bool], User], None, None]:
    username_base = get_random_name("user")
    created_users: List[User] = []

    def _make_user(
        level: AccessLevel = AccessLevel.DEVELOPER, add_to_project: bool = True
    ) -> User:
        last_id = len(created_users) + 1
        username = f"{username_base}_{last_id}"
        user = gitlab.create_user(
            username + "@example.com",
            username + " Example",
            username,
            get_random_password(),
        )
        user_obj = User(username, user["id"])
        if add_to_project:
            gitlab.add_member_to_project(
                group_and_project, None, level.value, user_id=user["id"]
            )
        created_users.append(user_obj)
        return user_obj

    yield _make_user

    for user in created_users:
        gitlab.delete_user(None, user_id=user.id)


@pytest.fixture(scope="class")
def three_members(gitlab, group_and_project, make_user):
    member1 = make_user()
    member2 = make_user()
    member3 = make_user()

    yield [member1.name, member2.name, member3.name]


@pytest.fixture(scope="function")
def outsider_user(gitlab, group_and_project):
    username = get_random_name("outsider_user")
    user = gitlab.create_user(
        username + "@example.com",
        username + " Example",
        username,
        get_random_password(),
    )
    user_obj = User(username, user["id"])

    yield user_obj.name

    gitlab.delete_user(None, user_id=user_obj.id)


@pytest.fixture(scope="function")
def public_ssh_key(gitlab):
    key = rsa.generate_private_key(
        backend=crypto_default_backend(), public_exponent=65537, key_size=2048
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH
    )

    yield public_key.decode("UTF-8")


@pytest.fixture(scope="function")
def other_public_ssh_key(gitlab):
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    key = rsa.generate_private_key(
        backend=crypto_default_backend(), public_exponent=65537, key_size=2048
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH
    )

    yield public_key.decode("UTF-8")


@pytest.fixture(scope="class")
def token_from_env_var():
    token = os.environ["GITLAB_TOKEN"]
    del os.environ["GITLAB_TOKEN"]

    yield token

    os.environ["GITLAB_TOKEN"] = token



from gitlabform.gitlab.groups import GitLabGroups


class GitLabGroupBadges(GitLabGroups):
    def get_group_badges(self, group_path):
        # (unlike the one for project badges) this endpoint returns ONLY group badges
        return self._make_requests_to_api(
            "groups/%s/badges",
            group_path,
            expected_codes=200,
        )

    def add_group_badge(
        self,
        group_path,
        badge_in_config,
    ):
        return self._make_requests_to_api(
            "groups/%s/badges",
            group_path,
            method="POST",
            data=badge_in_config,
            expected_codes=201,
        )

    def edit_group_badge(
        self,
        group_path,
        badge_in_gitlab,
        badge_in_config,
    ):
        return self._make_requests_to_api(
            "groups/%s/badges/%s",
            (group_path, badge_in_gitlab["id"]),
            method="PUT",
            data=badge_in_config,
        )

    def delete_group_badge(
        self,
        group_path,
        badge_in_gitlab,
    ):
        # 404 means it is already removed, so let's accept it for idempotency
        return self._make_requests_to_api(
            "groups/%s/badges/%s",
            (group_path, badge_in_gitlab["id"]),
            method="DELETE",
            expected_codes=[200, 204, 404],
        )



from gitlabform.gitlab.core import GitLabCore


class GitLabUsers(GitLabCore):
    def create_user(self, email, name, username, password):
        data = {
            "email": email,
            "name": name,
            "username": username,
            "password": password,
        }
        return self._make_requests_to_api(
            "users", method="POST", data=data, expected_codes=201
        )

    def get_user_by_name(self, username, user_id=None):
        if not user_id:
            user_id = self._get_user_id(username)
        return self._make_requests_to_api("users/%s", str(user_id), "GET")

    def delete_user(self, username, user_id=None):
        if not user_id:
            user_id = self._get_user_id(username)
        # 404 means that the user does not exist anymore, so let's accept it for idempotency
        return self._make_requests_to_api(
            "users/%s", str(user_id), method="DELETE", expected_codes=[204, 404]
        )



from abc import ABC, abstractmethod

from cli_ui import fatal

from gitlabform.constants import EXIT_INVALID_INPUT
from gitlabform.lists import OmissionReason, Groups, Projects


# Groups and projects filters jobs is to omit some groups and projects that GitLabForm is requested
# to process for a speed-up.
#
# An example reason for omitting groups and projects is when they have an empty effective config.


class GroupsAndProjectsFilters:
    def __init__(self, configuration, group_processors, project_processors):
        self.configuration = configuration

        self.omit_empty_configs = OmitEmptyConfigs(
            configuration, group_processors, project_processors
        )
        # add next filters here

    def filter(self, groups: Groups, projects: Projects):
        self.omit_empty_configs.filter(groups, projects)
        # add next filters here


class GroupsAndProjectsFilter(ABC):
    @abstractmethod
    def filter(self, groups: Groups, projects: Projects):
        pass


class OmitEmptyConfigs(GroupsAndProjectsFilter):
    """
    One of the reasons groups or project can be omitted from processing is when they have an empty effective config.

    For example with a config like:

    projects_and_groups:
      *:
        group_variables:
          foobar:
            key: "a key"
            value: "the value"

    ...and a request query "foo/bar" that points to a project "bar" in a group "foo", the effective config
    is empty, as "group_variables" is a group-level configuration and "foo/bar" is a project.
    """

    def __init__(self, configuration, group_processors, project_processors):
        self.configuration = configuration
        self.group_processors = group_processors
        self.project_processors = project_processors

        if not self.configuration.get("projects_and_groups", {}):
            fatal(
                "Configuration has to contain non-empty 'projects_and_groups' key.",
                exit_code=EXIT_INVALID_INPUT,
            )

    def filter(self, groups: Groups, projects: Projects) -> None:
        """
        :param groups: list of groups (and possibly subgroups)
        :param projects: list of projects
        """

        groups_with_empty_configs = []
        for group in groups.get_effective():
            if self._group_has_empty_effective_config(group):
                groups_with_empty_configs.append(group)
        groups.add_omitted(OmissionReason.EMPTY, groups_with_empty_configs)

        projects_with_empty_configs = []
        for project in projects.get_effective():
            if self._project_has_empty_effective_config(project):
                projects_with_empty_configs.append(project)
        projects.add_omitted(OmissionReason.EMPTY, projects_with_empty_configs)

    def _group_has_empty_effective_config(self, group: str) -> bool:
        """
        :param group: group/subgroup
        :return: if given group/subgroup has no config that can be processed
                 by any group-level processors
        """
        config_for_group = self.configuration.get_effective_config_for_group(group)
        for configuration_name in config_for_group.keys():
            if configuration_name in self.group_processors.get_configuration_names():
                return False
        return True

    def _project_has_empty_effective_config(self, project: str) -> bool:
        """
        :param project: 'group/project'
        :return: if given project has no config that can be processed
                 by any project-level processors
        """
        config_for_project = self.configuration.get_effective_config_for_project(
            project
        )
        for configuration_name in config_for_project.keys():
            if configuration_name in self.project_processors.get_configuration_names():
                return False
        return True



from logging import debug
from cli_ui import debug as verbose

import abc
from typing import Callable, Optional

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.util.difference_logger import DifferenceLogger


def noop():
    pass


class SingleEntityProcessor(AbstractProcessor, metaclass=abc.ABCMeta):
    def __init__(
        self,
        configuration_name: str,
        gitlab: GitLab,
        get_method_name: str,
        edit_method_name: str,
        add_method_name: Optional[str] = None,
    ):
        super().__init__(configuration_name, gitlab)
        self.get_method: Callable = getattr(self.gitlab, get_method_name)
        self.edit_method: Callable = getattr(self.gitlab, edit_method_name)
        if add_method_name:
            self.add_method: Callable = getattr(self.gitlab, add_method_name)
        else:
            self.add_method = noop

    def _process_configuration(self, project_or_group: str, configuration: dict):

        entity_config = configuration[self.configuration_name]

        entity_in_gitlab = self.get_method(project_or_group)
        debug(f"{self.configuration_name} BEFORE: ^^^")

        if entity_in_gitlab:
            if self._needs_update(entity_in_gitlab, entity_config):
                verbose(f"Editing {self.configuration_name} in {project_or_group}")
                self.edit_method(project_or_group, entity_config)
                debug(f"{self.configuration_name} AFTER: ^^^")
            else:
                verbose(
                    f"{self.configuration_name} in {project_or_group} doesn't need an update."
                )
        else:
            verbose(f"Adding {self.configuration_name} in {project_or_group}")
            self.add_method(project_or_group, entity_config)
            debug(f"{self.configuration_name} AFTER: ^^^")

    def _print_diff(self, project_or_project_and_group: str, entity_config):

        entity_in_gitlab = self.get_method(project_or_project_and_group)

        DifferenceLogger.log_diff(
            f"{self.configuration_name} changes",
            entity_in_gitlab,
            entity_config,
        )



from gitlabform.gitlab.core import GitLabCore


class GitLabCommits(GitLabCore):
    def get_commit(self, project_and_group_name, sha):
        return self._make_requests_to_api(
            "projects/%s/repository/commits/%s", (project_and_group_name, sha)
        )

    def get_ahead_and_behind(
        self, project_and_group_name, protected_branch, feature_branch
    ):
        ahead = 0
        behind = 0

        response = self._make_requests_to_api(
            "projects/%s/repository/compare?from=%s&to=%s",
            (project_and_group_name, protected_branch, feature_branch),
        )
        if len(response) > 0:
            ahead = len(response["commits"])

        response = self._make_requests_to_api(
            "projects/%s/repository/compare?from=%s&to=%s",
            (project_and_group_name, feature_branch, protected_branch),
        )
        if len(response) > 0:
            behind = len(response["commits"])

        return ahead, behind

    def get_last_commit(self, project_and_group_name, branch_name):
        branch = self._make_requests_to_api(
            "projects/%s/repository/branches/%s", (project_and_group_name, branch_name)
        )
        last_commit_hash = branch["commit"]["id"]
        return self.get_commit(project_and_group_name, last_commit_hash)

    def get_last_commit_attributes(self, project_and_group_name, branch):
        commit = self.get_last_commit(project_and_group_name, branch)

        # we want to read git's *commit date* instead of *author date*, so that "touching" the branch will invalidate
        # its "getting older" counter

        return commit["author_name"], commit["author_email"], commit["committed_date"]



from tests.acceptance import (
    run_gitlabform,
)


def single_true(iterable):
    i = iter(iterable)
    # use the fact that thanks to iter() we will "consume"
    # all the elements of iterable up to the one that is True
    # so that there should be no other True in the rest
    return any(i) and not any(i)


class TestDeployKeys:
    def test__deploy_key_add(
        self, gitlab, group_and_project_for_function, public_ssh_key
    ):

        config_add = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {public_ssh_key}
                title: some_key
        """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

    def test__deploy_key_add_delete(
        self, gitlab, group_and_project_for_function, public_ssh_key
    ):

        config_add = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {public_ssh_key}
                title: some_key
        """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

        config_delete = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                title: some_key
                delete: true
        """
        run_gitlabform(config_delete, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["title"] == "some_key" for key in deploy_keys])

    def test__deploy_key_add_delete_with_enforce(
        self, gitlab, group_and_project_for_function, public_ssh_key
    ):

        config_add = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {public_ssh_key}
                title: some_key
        """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert len(deploy_keys) >= 1
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

        config_delete = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              enforce: true
        """
        run_gitlabform(config_delete, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert len(deploy_keys) == 0

    def test__deploy_key_add_delete_readd(
        self, gitlab, group_and_project_for_function, public_ssh_key
    ):

        config_add = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {public_ssh_key}
                title: some_key
        """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

        config_delete = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                title: some_key
                delete: true
        """
        run_gitlabform(config_delete, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["title"] == "some_key" for key in deploy_keys])

        run_gitlabform(config_add, group_and_project_for_function)
        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "some_key" for key in deploy_keys])

    def test__deploy_key_add_delete_readd_under_different_name(
        self, gitlab, group_and_project_for_function, public_ssh_key
    ):

        config_add = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    key: {public_ssh_key}
                    title: a_title
            """
        run_gitlabform(config_add, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "a_title" for key in deploy_keys])

        config_delete = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    title: a_title
                    delete: true
            """
        run_gitlabform(config_delete, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["title"] == "a_title" for key in deploy_keys])

        config_readd = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    key: {public_ssh_key}
                    title: another_title
            """
        run_gitlabform(config_readd, group_and_project_for_function)
        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "another_title" for key in deploy_keys])

    def test__deploy_key_update_title(
        self, gitlab, group_and_project_for_function, public_ssh_key
    ):

        config = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  foobar:
                    key: {public_ssh_key}
                    title: title_before
            """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "title_before" for key in deploy_keys])

        config = f"""
            projects_and_groups:
              {group_and_project_for_function}:
                deploy_keys:
                  # because it's the title that defines the key, we cannot immediately change key's title in the config.
                  # we need two element config - 1. delete it under old title, 2. add it under new title
                  foo:
                    title: title_before
                    delete: true
                  bar:
                    key: {public_ssh_key}
                    title: title_after
            """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert single_true([key["title"] == "title_after" for key in deploy_keys])
        assert not any([key["title"] == "title_before" for key in deploy_keys])

    def test__deploy_key_update_value(
        self,
        gitlab,
        group_and_project_for_function,
        public_ssh_key,
        other_public_ssh_key,
    ):

        config = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {public_ssh_key}
                title: a_key
        """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert any([key["key"] == public_ssh_key for key in deploy_keys])

        config = f"""
        projects_and_groups:
          {group_and_project_for_function}:
            deploy_keys:
              foobar:
                key: {other_public_ssh_key}
                title: a_key
        """
        run_gitlabform(config, group_and_project_for_function)

        deploy_keys = gitlab.get_deploy_keys(group_and_project_for_function)
        assert not any([key["key"] == public_ssh_key for key in deploy_keys])
        assert any([key["key"] == other_public_ssh_key for key in deploy_keys])



from logging import debug
from typing import Dict, Tuple

from cli_ui import fatal

from gitlabform.constants import EXIT_INVALID_INPUT
from gitlabform.gitlab import GitLab, AccessLevel
from gitlabform.processors.abstract_processor import AbstractProcessor


class GroupMembersProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("group_members", gitlab)

    def _process_configuration(self, group: str, configuration: dict):

        enforce_group_members = configuration.get("group_members|enforce", False)

        (
            groups_to_set_by_group_path,
            users_to_set_by_username,
        ) = self._get_groups_and_users_to_set(configuration)

        if (
            enforce_group_members
            and not groups_to_set_by_group_path
            and not users_to_set_by_username
        ):
            fatal(
                "Group members configuration section has to contain"
                " some 'users' or 'groups' defined as Owners,"
                " if you want to enforce them (GitLab requires it).",
                exit_code=EXIT_INVALID_INPUT,
            )

        self._process_groups(group, groups_to_set_by_group_path, enforce_group_members)

        self._process_users(group, users_to_set_by_username, enforce_group_members)

    @staticmethod
    def _get_groups_and_users_to_set(configuration: dict) -> Tuple[dict, dict]:

        groups_to_set_by_group_path = configuration.get("group_members|groups", {})

        users_to_set_by_username = configuration.get("group_members", {})
        if users_to_set_by_username:
            proper_users_to_set_by_username = configuration.get(
                "group_members|users", {}
            )
            if proper_users_to_set_by_username:
                users_to_set_by_username = proper_users_to_set_by_username
            else:
                users_to_set_by_username.pop("enforce", None)
                users_to_set_by_username.pop("users", None)
                users_to_set_by_username.pop("groups", None)

        return groups_to_set_by_group_path, users_to_set_by_username

    def _process_groups(
        self, group: str, groups_to_set_by_group_path: dict, enforce_group_members: bool
    ):

        # group users before by group name
        groups_before = self.gitlab.get_group_case_insensitive(group)[
            "shared_with_groups"
        ]
        debug("Group shared with BEFORE: %s", groups_before)

        groups_before_by_group_path = dict()
        for share_details in groups_before:
            groups_before_by_group_path[
                share_details["group_full_path"]
            ] = share_details

        for share_with_group_path in groups_to_set_by_group_path:

            group_access_to_set = groups_to_set_by_group_path[share_with_group_path][
                "group_access"
            ]

            expires_at_to_set = (
                groups_to_set_by_group_path[share_with_group_path]["expires_at"]
                if "expires_at" in groups_to_set_by_group_path[share_with_group_path]
                else None
            )

            if share_with_group_path in groups_before_by_group_path:

                group_access_before = groups_before_by_group_path[
                    share_with_group_path
                ]["group_access_level"]
                expires_at_before = groups_before_by_group_path[share_with_group_path][
                    "expires_at"
                ]

                if (
                    group_access_before == group_access_to_set
                    and expires_at_before == expires_at_to_set
                ):
                    debug(
                        "Nothing to change for group '%s' - same config now as to set.",
                        share_with_group_path,
                    )
                else:
                    debug(
                        "Re-adding group '%s' to change their access level or expires at.",
                        share_with_group_path,
                    )
                    # we will remove the group first and then re-add them,
                    # to ensure that the group has the expected access level
                    self.gitlab.remove_share_from_group(group, share_with_group_path)
                    self.gitlab.add_share_to_group(
                        group,
                        share_with_group_path,
                        group_access_to_set,
                        expires_at_to_set,
                    )

            else:
                debug(
                    "Adding group '%s' who previously was not a member.",
                    share_with_group_path,
                )
                self.gitlab.add_share_to_group(
                    group, share_with_group_path, group_access_to_set, expires_at_to_set
                )

        if enforce_group_members:
            # remove groups not configured explicitly
            groups_not_configured = set(groups_before_by_group_path) - set(
                groups_to_set_by_group_path
            )
            for group_path in groups_not_configured:
                debug(
                    "Removing group '%s' who is not configured to be a member.",
                    group_path,
                )
                self.gitlab.remove_share_from_group(group, group_path)
        else:
            debug("Not enforcing group members.")

        debug("Group shared with AFTER: %s", self.gitlab.get_group_members(group))

    def _process_users(
        self, group: str, users_to_set_by_username: dict, enforce_group_members: bool
    ):

        # group users before by username
        # (note: we DON'T get inherited users as we don't manage them at this level anyway)
        users_before = self.gitlab.get_group_members(group, with_inherited=False)
        debug("Group members BEFORE: %s", users_before)
        users_before_by_username = dict()
        for user in users_before:
            users_before_by_username[user["username"]] = user

        if users_to_set_by_username:

            # group users to set by access level
            users_to_set_by_access_level: Dict[int, list] = dict()
            for user in users_to_set_by_username:
                access_level = users_to_set_by_username[user]["access_level"]
                users_to_set_by_access_level.setdefault(access_level, []).append(user)

            # we HAVE TO start configuring access from the highest access level - in case of groups this is Owner
            # - to ensure that we won't end up with no Owner in a group
            for level in reversed(sorted(AccessLevel.group_levels())):

                users_to_set_with_this_level = (
                    users_to_set_by_access_level[level]
                    if level in users_to_set_by_access_level
                    else []
                )

                for user in users_to_set_with_this_level:

                    access_level_to_set = users_to_set_by_username[user]["access_level"]
                    expires_at_to_set = (
                        users_to_set_by_username[user]["expires_at"]
                        if "expires_at" in users_to_set_by_username[user]
                        else None
                    )

                    if user in users_before_by_username:

                        access_level_before = users_before_by_username[user][
                            "access_level"
                        ]
                        expires_at_before = users_before_by_username[user]["expires_at"]

                        if (
                            access_level_before == access_level_to_set
                            and expires_at_before == expires_at_to_set
                        ):
                            debug(
                                "Nothing to change for user '%s' - same config now as to set.",
                                user,
                            )
                        else:
                            debug(
                                "Re-adding user '%s' to change their access level or expires at.",
                                user,
                            )
                            # we will remove the user first and then re-add they,
                            # to ensure that the user has the expected access level
                            self.gitlab.remove_member_from_group(group, user)
                            self.gitlab.add_member_to_group(
                                group, user, access_level_to_set, expires_at_to_set
                            )

                    else:
                        debug("Adding user '%s' who previously was not a member.", user)
                        self.gitlab.add_member_to_group(
                            group, user, access_level_to_set, expires_at_to_set
                        )

        if enforce_group_members:
            # remove users not configured explicitly
            # note: only direct members are removed - inherited are left
            users_not_configured = set(
                [user["username"] for user in users_before]
            ) - set(users_to_set_by_username.keys())
            for user in users_not_configured:
                debug("Removing user '%s' who is not configured to be a member.", user)
                self.gitlab.remove_member_from_group(group, user)
        else:
            debug("Not enforcing group members.")

        debug("Group members AFTER: %s", self.gitlab.get_group_members(group))



from abc import ABC

from gitlabform.configuration.core import ConfigurationCore


class ConfigurationCommon(ConfigurationCore, ABC):
    """
    Gets the common configuration, applied to all groups and projects.
    """

    def get_common_config(self) -> dict:
        """
        :return: literal common configuration or empty dict if not defined
        """
        common_config = self.get("projects_and_groups|*", {})
        if common_config:
            section_name = "*"
            self._validate_break_inheritance_flag(common_config, section_name)
        return common_config



from gitlabform.gitlab.core import GitLabCore, NotFoundException


class GitLabBranches(GitLabCore):
    def protect_branch(self, project_and_group_name, branch, protect_settings):

        url = "projects/%s/protected_branches?name=%s"
        parameters_list = [
            project_and_group_name,
            branch,
        ]
        return self._make_requests_to_api(
            url,
            tuple(parameters_list),
            method="POST",
            expected_codes=[
                200,
                201,
            ],
            json=protect_settings,
        )

    def set_branch_code_owner_approval_required(
        self,
        project_and_group_name,
        branch,
        code_owner_approval_required,
    ):
        data = {
            "id": project_and_group_name,
            "branch": branch,
            "code_owner_approval_required": code_owner_approval_required,
        }
        return self._make_requests_to_api(
            "projects/%s/protected_branches/%s",
            (
                project_and_group_name,
                branch,
            ),
            method="PATCH",
            data=data,
            expected_codes=[200, 201],
        )

    def unprotect_branch(self, project_and_group_name, branch):
        # 404 means that the branch is already unprotected
        return self._make_requests_to_api(
            "projects/%s/protected_branches/%s",
            (project_and_group_name, branch),
            method="DELETE",
            expected_codes=[200, 201, 204, 404],
        )

    def get_branches(self, project_and_group_name):
        result = self._make_requests_to_api(
            "projects/%s/repository/branches", project_and_group_name
        )
        return sorted(map(lambda x: x["name"], result))

    def get_branch(self, project_and_group_name, branch):
        return self._make_requests_to_api(
            "projects/%s/repository/branches/%s", (project_and_group_name, branch)
        )

    def get_branch_access_levels(self, project_and_group_name, branch):
        return self._make_requests_to_api(
            "projects/%s/protected_branches/%s", (project_and_group_name, branch)
        )

    def get_only_branch_access_levels(self, project_and_group_name, branch):
        try:
            result = self._make_requests_to_api(
                "projects/%s/protected_branches/%s", (project_and_group_name, branch)
            )

            push_access_levels = set()
            merge_access_levels = set()
            push_access_user_ids = set()
            merge_access_user_ids = set()
            unprotect_access_level = None

            if "push_access_levels" in result:
                for push_access in result["push_access_levels"]:
                    if not push_access["user_id"]:
                        push_access_levels.add(push_access["access_level"])
                    else:
                        push_access_user_ids.add(push_access["user_id"])

            if "merge_access_levels" in result:
                for merge_access in result["merge_access_levels"]:
                    if not merge_access["user_id"]:
                        merge_access_levels.add(merge_access["access_level"])
                    else:
                        merge_access_user_ids.add(merge_access["user_id"])

            if (
                "unprotect_access_levels" in result
                and len(result["unprotect_access_levels"]) == 1
            ):
                unprotect_access_level = result["unprotect_access_levels"][0][
                    "access_level"
                ]

            return (
                sorted(push_access_levels),
                sorted(merge_access_levels),
                sorted(push_access_user_ids),
                sorted(merge_access_user_ids),
                unprotect_access_level,
            )
        except NotFoundException:
            return None, None, None, None, None

    def create_branch(
        self, project_and_group_name, new_branch_name, create_branch_from_ref
    ):
        data = {
            "id": project_and_group_name,
            "branch": new_branch_name,
            "ref": create_branch_from_ref,
        }
        self._make_requests_to_api(
            "projects/%s/repository/branches",
            project_and_group_name,
            method="POST",
            data=data,
            expected_codes=[200, 201],
        )

    def delete_branch(self, project_and_group_name, branch):
        self._make_requests_to_api(
            "projects/%s/repository/branches/%s",
            (project_and_group_name, branch),
            method="DELETE",
            expected_codes=[200, 204],
        )

    def get_protected_branches(self, project_and_group_name):
        protected_branches = self._make_requests_to_api(
            "projects/%s/protected_branches", project_and_group_name
        )
        return [branch["name"] for branch in protected_branches]

    def get_unprotected_branches(self, project_and_group_name):
        branches = self._make_requests_to_api(
            "projects/%s/repository/branches", project_and_group_name
        )

        unprotected_branches = []
        for branch in branches:
            if not branch["protected"]:
                name = branch["name"]
                unprotected_branches.append(name)

        return unprotected_branches



from gitlabform.gitlab import GitLab
from gitlabform.processors.single_entity_processor import SingleEntityProcessor


class ProjectPushRulesProcessor(SingleEntityProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "project_push_rules",
            gitlab,
            get_method_name="get_project_push_rules",
            edit_method_name="put_project_push_rules",
            add_method_name="post_project_push_rules",
        )



# f.e. bad syntax in the config file ~= "it's your fault" 😅
EXIT_INVALID_INPUT = 1
# f.e. when requests to GitLab fail ~= "it's not your fault" 😎
EXIT_PROCESSING_ERROR = 2

# legacy single approval rule name
APPROVAL_RULE_NAME = "Approvers (configured using GitLabForm)"



import textwrap


from gitlabform.processors.util.difference_logger import DifferenceLogger


def test_empty_dict_current():
    current_config = dict()
    config_to_apply = {
        "foo": 123,
        "bar": "whatever",
    }
    result = DifferenceLogger.log_diff(
        "test", current_config, config_to_apply, False, None, True
    )
    # the whitespace after "123" below is required!
    expected = textwrap.dedent(
        """
        test:
        foo: "???" => 123       
        bar: "???" => "whatever"
    """
    ).strip()
    assert result == expected


def test_none_current():
    current_config = None
    config_to_apply = {
        "foo": 123,
        "bar": "whatever",
    }
    result = DifferenceLogger.log_diff(
        "test", current_config, config_to_apply, False, None, True
    )
    # the whitespace after "123" below is required!
    expected = textwrap.dedent(
        """
        test:
        foo: "???" => 123       
        bar: "???" => "whatever"
    """
    ).strip()
    assert result == expected
